<?php

namespace Oro\Bundle\EmailBundle\Form\Type;

use Doctrine\Common\Util\ClassUtils;

use Symfony\Component\Form\AbstractType;
use Symfony\Component\Form\FormBuilderInterface;
use Symfony\Component\OptionsResolver\OptionsResolverInterface;

use Oro\Bundle\EmailBundle\Form\DataTransformer\OriginTransformer;
use Oro\Bundle\EmailBundle\Builder\Helper\EmailModelBuilderHelper;
use Oro\Bundle\EmailBundle\Tools\EmailOriginHelper;
use Oro\Bundle\EntityBundle\ORM\Registry;
use Oro\Bundle\EmailBundle\Provider\RelatedEmailsProvider;
use Oro\Bundle\ImapBundle\Entity\UserEmailOrigin;
use Oro\Bundle\SecurityBundle\SecurityFacade;
use Oro\Bundle\UserBundle\Entity\User;
use Oro\Bundle\EmailBundle\Entity\Manager\MailboxManager;

class EmailOriginFromType extends AbstractType
{
    const NAME = 'oro_email_email_origin_from';

    /** @var SecurityFacade */
    protected $securityFacade;

    /** @var EmailModelBuilderHelper */
    protected $helper;

    /** @var RelatedEmailsProvider */
    protected $relatedEmailsProvider;

    /** @var MailboxManager */
    protected $mailboxManager;

    /** @var Registry */
    protected $registry;

    /** @var EmailOriginHelper */
    protected $emailOriginHelper;

    /**
     * @param SecurityFacade $securityFacade
     * @param RelatedEmailsProvider $relatedEmailsProvider
     * @param EmailModelBuilderHelper $helper
     * @param MailboxManager $mailboxManager
     * @param Registry $registry
     * @param EmailOriginHelper $emailOriginHelper
     */
    public function __construct(
        SecurityFacade $securityFacade,
        RelatedEmailsProvider $relatedEmailsProvider,
        EmailModelBuilderHelper $helper,
        MailboxManager $mailboxManager,
        Registry $registry,
        EmailOriginHelper $emailOriginHelper
    ) {
        $this->securityFacade = $securityFacade;
        $this->relatedEmailsProvider = $relatedEmailsProvider;
        $this->helper = $helper;
        $this->mailboxManager = $mailboxManager;
        $this->registry = $registry;
        $this->emailOriginHelper = $emailOriginHelper;
    }

    /**
     * {@inheritdoc}
     */
    public function buildForm(FormBuilderInterface $builder, array $options)
    {
        $builder->addModelTransformer(
            new OriginTransformer(
                $this->registry->getManager(),
                $this->securityFacade,
                $this->emailOriginHelper
            )
        );
    }

    /**
     * {@inheritdoc}
     */
    public function setDefaultOptions(OptionsResolverInterface $resolver)
    {
        $choices = $this->createChoices();
        $resolver->setDefaults([
            'choices'   => $choices,
            'read_only' => count($choices) === 1,
        ]);
    }

    /**
     * @return array
     */
    protected function createChoices()
    {
        $user = $this->securityFacade->getLoggedUser();
        if (!$user instanceof User) {
            return [];
        }
        $origins = [];
        $origins = $this->fillUserOrigins($user, $origins);
        if (count($origins) === 0) {
            $origins = $this->fillUserEmails($user, $origins);
        }
        $origins = $this->fillMailboxOrigins($user, $origins);

        return $origins;
    }

    /**
     * {@inheritdoc}
     */
    public function getParent()
    {
        return 'genemu_jqueryselect2_choice';
    }

    /**
     * {@inheritdoc}
     */
    public function getName()
    {
        return $this->getBlockPrefix();
    }

    /**
     * {@inheritdoc}
     */
    public function getBlockPrefix()
    {
        return static::NAME;
    }

    /**
     * @param User $user
     * @param array $origins
     *
     * @return array
     */
    protected function fillUserOrigins(User $user, $origins)
    {
        $userOrigins = $user->getEmailOrigins();
        foreach ($userOrigins as $origin) {
            if ($origin instanceof UserEmailOrigin) {
                if ($origin->isActive()) {
                    $owner = $origin->getOwner();
                    $email = $origin->getOwner()->getEmail();
                    $this->helper->preciseFullEmailAddress($email, ClassUtils::getClass($owner), $owner->getId());
                    $origins[$origin->getId() . '|' . $origin->getOwner()->getEmail()] = $email;
                }
            }
        }
        return $origins;
    }

    /**
     * @param User $user
     * @param array $origins
     * @return array
     */
    protected function fillUserEmails(User $user, $origins)
    {
        $email = $user->getEmail();
        $origins = $this->processFillUserEmail($email, $origins, $user);

        $userEmails = $user->getEmails();
        foreach ($userEmails as $email) {
            $email = $email->getEmail();
            $origins = $this->processFillUserEmail($email, $origins);
        }

        return $origins;
    }

    /**
     * @param string $email
     * @param array $origins
     * @param $owner
     *
     * @return mixed
     */
    protected function processFillUserEmail($email, $origins, $owner = null)
    {
        $key = '0|' . $email;
        if (!array_key_exists($key, $origins)) {
            if ($owner) {
                $this->helper->preciseFullEmailAddress($email, ClassUtils::getClass($owner), $owner->getId());
            } else {
                $this->helper->preciseFullEmailAddress($email);
            }
            $origins[$key] = $email;
        }

        return $origins;
    }

    /**
     * @param User $user
     * @param array $origins
     *
     * @return mixed
     */
    protected function fillMailboxOrigins(User $user, $origins)
    {
        $mailboxes = $this->mailboxManager->findAvailableMailboxes($user, $this->securityFacade->getOrganization());
        foreach ($mailboxes as $mailbox) {
            $origin = $mailbox->getOrigin();
            if ($origin->isActive()) {
                $email = $mailbox->getEmail();
                $this->helper->preciseFullEmailAddress($email);
                $email .= ' (Mailbox)';
                $origins[$origin->getId() . '|' . $mailbox->getEmail()] = $email;
            }
        }

        return $origins;
    }
}
