<?php

namespace Oro\Bundle\NotificationBundle\Tests\Unit\Model;

use Doctrine\ORM\EntityManager;
use Oro\Bundle\ConfigBundle\Config\ConfigManager;
use Oro\Bundle\EmailBundle\Entity\Repository\EmailTemplateRepository;
use Oro\Bundle\EmailBundle\Model\EmailTemplateInterface;
use Oro\Bundle\LocaleBundle\DQL\DQLNameFormatter;
use Oro\Bundle\NotificationBundle\Doctrine\EntityPool;
use Oro\Bundle\NotificationBundle\Manager\EmailNotificationManager;
use Oro\Bundle\NotificationBundle\Model\MassNotification;
use Oro\Bundle\NotificationBundle\Model\MassNotificationSender;
use Oro\Bundle\UserBundle\Entity\Repository\UserRepository;

class MassNotificationSenderTest extends \PHPUnit_Framework_TestCase
{
    const TEST_SENDER_EMAIL = 'admin@example.com';
    const TEST_SENDER_NAME  = 'sender name';
    const TEMPLATE_NAME     = 'test template';

    /** @var \PHPUnit_Framework_MockObject_MockObject */
    protected $entityManager;

    /** @var \PHPUnit_Framework_MockObject_MockObject */
    protected $userRepository;

    /** @var \PHPUnit_Framework_MockObject_MockObject */
    protected $templateRepository;

    /** @var \PHPUnit_Framework_MockObject_MockObject */
    protected $entityPool;

    /** @var \PHPUnit_Framework_MockObject_MockObject|ConfigManager */
    protected $cm;

    /** @var \PHPUnit_Framework_MockObject_MockObject|EmailNotificationManager */
    protected $manager;

    /** @var \PHPUnit_Framework_MockObject_MockObject */
    protected $dqlNameFormatter;

    /** @var MassNotificationSender */
    protected $sender;

    /** @var  array */
    protected $massNotificationParams;

    protected function setUp()
    {
        $this->entityManager = $this->getMockBuilder(EntityManager::class)
            ->disableOriginalConstructor()->setMethods(['getRepository'])->getMock();

        $this->userRepository = $this->getMockBuilder(UserRepository::class)
                    ->disableOriginalConstructor()->getMock();

        $this->templateRepository =
            $this->getMockBuilder(EmailTemplateRepository::class)
                    ->disableOriginalConstructor()->getMock();

        $this->entityPool = $this->getMockBuilder(EntityPool::class)
            ->disableOriginalConstructor()->getMock();

        $this->cm = $this->getMockBuilder(ConfigManager::class)
            ->disableOriginalConstructor()->getMock();

        $this->dqlNameFormatter = $this->getMockBuilder(DQLNameFormatter::class)
            ->disableOriginalConstructor()->getMock();

        $this->manager = $this->getMockBuilder(EmailNotificationManager::class)
            ->disableOriginalConstructor()
            ->getMock();

        $this->sender = new MassNotificationSender(
            $this->manager,
            $this->cm,
            $this->entityManager,
            $this->entityPool,
            $this->dqlNameFormatter
        );
    }

    protected function tearDown()
    {
        unset($this->entityManager);
        unset($this->cm);
        unset($this->entityPool);
        unset($this->manager);
        unset($this->sender);
        unset($this->templateRepository);
        unset($this->userRepository);
        unset($this->dqlNameFormatter);
    }

    public function testSendToActiveUsersWithEmptySender()
    {
        $body = "Test Body";
        $subject = "Test Subject";
        $userRecipients = [
            ['email' => 'test1@test.com', 'name' => 'test1'],
            ['email' => 'test2@test.com', 'name' => 'test2']
        ];
        $this->cm->expects($this->any())->method('get')->will(
            $this->returnValueMap(
                [
                    ['oro_notification.email_notification_sender_email', false, false, null, self::TEST_SENDER_EMAIL],
                    ['oro_notification.email_notification_sender_name', false, false, null, self::TEST_SENDER_NAME],
                    ['oro_notification.mass_notification_recipients', false, false, null, ''],
                    ['oro_notification.mass_notification_template', false, false, null, self::TEMPLATE_NAME]
                ]
            )
        );

        $this->entityManager->expects($this->at(1))->method('getRepository')->with('OroEmailBundle:EmailTemplate')
            ->will($this->returnValue($this->templateRepository));

        $this->entityManager->expects($this->at(0))->method('getRepository')->with('OroUserBundle:User')->will(
            $this->returnValue($this->userRepository)
        );

        $template = $this->createMock('Oro\Bundle\EmailBundle\Entity\EmailTemplate');
        $template->expects($this->once())->method('getType')->will($this->returnValue('html'));
        $template->expects($this->once())->method('getContent')->will($this->returnValue('test content'));
        $template->expects($this->once())->method('getSubject')->will($this->returnValue('subject'));

        $this->templateRepository->expects($this->once())->method('findByName')->with(self::TEMPLATE_NAME)->will(
            $this->returnValue($template)
        );
        $this->dqlNameFormatter->expects($this->once())->method('getFormattedNameDQL')->will(
            $this->returnValue("ConcatExpression")
        );

        $query = $this->getMockBuilder('Doctrine\ORM\AbstractQuery')
            ->disableOriginalConstructor()
            ->setMethods(['getResult'])
            ->getMockForAbstractClass();
        $query->expects($this->any())
            ->method('getResult')
            ->will($this->returnValue($userRecipients));

        $queryBuilder = $this->getMockBuilder('Doctrine\ORM\QueryBuilder')
            ->disableOriginalConstructor()
            ->setMethods(['getQuery', 'andWhere', 'setParameter'])
            ->getMock();
        $queryBuilder->expects($this->once())
            ->method('andWhere')
            ->with('u.enabled = :enabled')
            ->will($this->returnSelf());
        $queryBuilder->expects($this->once())
            ->method('setParameter')
            ->with('enabled', true);
        $queryBuilder->expects($this->once())
            ->method('getQuery')
            ->will($this->returnValue($query));
        $this->userRepository->expects($this->once())->method('getPrimaryEmailsQb')->with("ConcatExpression")->will(
            $this->returnValue($queryBuilder)
        );

        $this->massNotificationParams = [
            'sender_name'      => self::TEST_SENDER_NAME,
            'sender_email'     => self::TEST_SENDER_EMAIL,
            'recipients'       => [['test1@test.com' => 'test1'], ['test2@test.com' => 'test2']],
            'template_type'    => 'html',
            'template_content' => 'test content',
            'template_subject' => $subject
        ];
        $this->manager->expects($this->once())->method('process')->with(
            null,
            $this->callback([$this, 'assertMassNotification']),
            null,
            [MassNotificationSender::MAINTENANCE_VARIABLE => $body]
        );

        $this->entityPool->expects($this->once())->method('persistAndFlush')->with($this->entityManager);

        $this->assertEquals(2, $this->sender->send($body, $subject));
    }

    public function testSendToConfigEmailsWithEmtpyTemplate()
    {
        $body = "Test Body";
        $subject = null;
        $senderName = "Sender Name";
        $senderEmail = "sender@test.com";
        $configRecipients = 'test1@test.com;test2@test.com';
        $this->cm->expects($this->any())->method('get')->will(
            $this->returnValueMap(
                [
                    ['oro_notification.mass_notification_recipients', false, false, null, $configRecipients],
                    ['oro_notification.mass_notification_template', false, false, null, self::TEMPLATE_NAME]
                ]
            )
        );

        $this->entityManager->expects($this->at(0))->method('getRepository')->with('OroEmailBundle:EmailTemplate')
                            ->will($this->returnValue($this->templateRepository));

        $this->templateRepository->expects($this->once())->method('findByName')->with(self::TEMPLATE_NAME)->will(
            $this->returnValue(null)
        );
        $this->massNotificationParams = [
            'sender_name'      => $senderName,
            'sender_email'     => $senderEmail,
            'recipients'       => explode(';', $configRecipients),
            'template_type'    => 'txt',
            'template_content' => sprintf("{{ %s }}", MassNotificationSender::MAINTENANCE_VARIABLE),
            'template_subject' => $subject
        ];
        $this->manager->expects($this->once())->method('process')->with(
            null,
            $this->callback([$this, 'assertMassNotification']),
            null,
            [MassNotificationSender::MAINTENANCE_VARIABLE => $body]
        );

        $this->assertEquals(2, $this->sender->send($body, $subject, $senderEmail, $senderName));
    }

    /**
     * @param array $massNotifications
     * @return bool
     */
    public function assertMassNotification($massNotifications)
    {
        $params = $this->massNotificationParams;
        /** @var MassNotification $massNotification */
        $massNotification = current($massNotifications);
        $template = $massNotification->getTemplate();
        $this->assertEquals($params['sender_name'], $massNotification->getSenderName());
        $this->assertEquals($params['sender_email'], $massNotification->getSenderEmail());
        $this->assertEquals($params['recipients'], $massNotification->getRecipientEmails());
        $this->assertTrue($template instanceof EmailTemplateInterface);
        $this->assertEquals($params['template_type'], $template->getType());
        $this->assertEquals($params['template_content'], $template->getContent());
        $this->assertEquals($params['template_subject'], $template->getSubject());

        return true;
    }
}
